@@ -24,9 +24,8 @@ suite() {
 }
 
 oneTimeSetUp() {
-	cd "$TESTDIR/../"
+	GIT_FTP_CMD="$(dirname "$TESTDIR")/git-ftp"
 
-	GIT_FTP_CMD="$(pwd)/git-ftp"
 	: ${GIT_FTP_USER=ftp}
 	: ${GIT_FTP_PASSWD=}
 	: ${GIT_FTP_ROOT=localhost}
@@ -89,7 +88,7 @@ test_displays_usage() {
 
 test_prints_version() {
 	version=$($GIT_FTP_CMD 2>&1 --version)
-	assertEquals = "git-ftp version 1.1.0"  "$version"
+	assertEquals = "git-ftp version 1.2.0"  "$version"
 }
 
 test_inits() {
@@ -148,14 +147,13 @@ test_push_unknown_commit() {
 }
 
 test_push_nothing() {
-	cd $GIT_PROJECT_PATH
 	init=$($GIT_FTP init)
 	# make some changes
 	echo "1" >> "./test 1.txt"
 	git commit -a -m "change" > /dev/null 2>&1
 	push=$($GIT_FTP push --dry-run)
 	assertEquals 0 $?
-	assertTrue "$push" "echo \"$push\" | grep 'There are 1 files to sync:'"
+	assertTrue "$push" "echo \"$push\" | grep '1 file to sync:'"
 	echo 'test 1.txt' >> .git-ftp-ignore
 	push=$($GIT_FTP push)
 	assertEquals 0 $?
@@ -334,6 +332,15 @@ test_scopes_uses_password_by_cli() {
 	assertEquals 0 $rtrn
 }
 
+test_scopes_add() {
+	add=$($GIT_FTP_CMD add-scope xyz ftpes://user:password@ftp.example.com/xyz)
+	assertEquals 0 $?
+	assertFalse 'Does not display warning with valid URL' "echo $add | grep ^Warning -q"
+	add=$($GIT_FTP_CMD add-scope xyz ftpes://user:pass:word@ftp.example.com/xyz)
+	assertEquals 0 $?
+	assertTrue 'Does display warning invalid URL' "echo $add | grep ^Warning -q"
+}
+
 test_delete() {
 	cd $GIT_PROJECT_PATH
 
@@ -537,6 +544,16 @@ test_include_ignore_push() {
 	assertFalse ' .htaccess.prod was uploaded' "remote_file_exists '.htaccess.prod'"
 }
 
+# Testing Github issue #245
+test_include_ignore_all_push() {
+	init=$($GIT_FTP init)
+	echo 'always include me' > untracked.txt
+	echo '!untracked.txt' > .git-ftp-include
+	git commit --allow-empty -m 'There are no changed files in this commit.' -q
+	push=$($GIT_FTP push)
+	assertTrue ' include file always' "remote_file_exists 'untracked.txt'"
+}
+
 test_include_ftp_ignore_init() {
 	cd $GIT_PROJECT_PATH
 	echo 'htaccess' > .htaccess
@@ -610,6 +627,22 @@ test_file_with_nonchar() {
 	assertFalse "file $file2 still exists in $CURL_URL" "remote_file_exists '$file2enc'"
 }
 
+# issue #209
+test_file_with_unicode() {
+	supports_unicode || startSkipping
+	file1='umlaut_ä.md'
+	file1enc='umlaut_%C3%A4.md'
+	echo 'content' > "$file1"
+	git add .
+	git commit -a -m 'added special filenames' -q
+	init=$($GIT_FTP init)
+	assertTrue " file $file1 not uploaded" "remote_file_equals '$file1' '$file1enc'"
+	git rm "$file1" -q
+	git commit -m 'delete' -q
+	push=$($GIT_FTP push)
+	assertFalse "file $file1 still exists in $CURL_URL" "remote_file_exists '$file1enc'"
+}
+
 test_syncroot() {
 	cd $GIT_PROJECT_PATH
 	syncroot='foo bar'
@@ -620,6 +653,171 @@ test_syncroot() {
 	assertTrue 'test failed: syncroot.txt not there as expected' "remote_file_exists 'syncroot.txt'"
 }
 
+test_download() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' > external.txt
+	curl -T external.txt $CURL_URL/ 2> /dev/null
+	rtrn=$?
+	assertEquals 0 $rtrn
+	rm external.txt
+	$GIT_FTP download > /dev/null 2>&1
+	rtrn=$?
+	assertEquals 0 $rtrn
+	assertTrue ' external file not downloaded' "[ -r 'external.txt' ]"
+}
+
+test_download_untracked() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	touch 'untracked.file'
+	$GIT_FTP download > /dev/null 2>&1
+	assertEquals 8 $?
+	assertFalse ' external file downloaded' "[ -f 'external.txt' ]"
+	assertTrue ' untracked file deleted' "[ -r 'untracked.file' ]"
+}
+
+test_download_syncroot() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	mkdir foobar && echo "test" > foobar/syncroot.txt
+	git add . > /dev/null 2>&1
+	git commit -a -m "syncroot test" > /dev/null 2>&1
+	init=$($GIT_FTP init --syncroot foobar)
+	echo 'foreign content' > external.txt
+	curl -T external.txt $CURL_URL/ 2> /dev/null
+	rm external.txt
+	$GIT_FTP download --syncroot foobar/ > /dev/null 2>&1
+	rtrn=$?
+	assertEquals 0 $rtrn
+	assertFalse ' external file downloaded to git root' "[ -r 'external.txt' ]"
+	assertTrue ' external file not downloaded to syncroot' "[ -r 'foobar/external.txt' ]"
+}
+
+test_download_dry_run() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	$GIT_FTP download --dry-run > /dev/null 2>&1
+	assertEquals 0 $?
+	assertTrue ' external file downloaded' "[ ! -e 'external.txt' ]"
+}
+
+test_pull() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' > external.txt
+	curl -T external.txt $CURL_URL/ 2> /dev/null
+	rm external.txt
+	echo 'own content' > internal.txt
+	git add . > /dev/null 2>&1
+	git commit -a -m "local modification" > /dev/null 2>&1
+	$GIT_FTP pull > /dev/null 2>&1
+	rtrn=$?
+	assertEquals 0 $rtrn
+	assertTrue ' external file not downloaded' "[ -r 'external.txt' ]"
+	assertTrue ' external file not listed in commit message' "git log | grep 'external.txt'"
+}
+
+test_pull_nothing() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	$GIT_FTP pull > /dev/null 2>&1
+	assertEquals 0 $?
+}
+
+test_pull_branch() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' > external.txt
+	curl -T external.txt $CURL_URL/ 2> /dev/null
+	rm external.txt
+	echo 'own content' > internal.txt
+	git add . > /dev/null 2>&1
+	git commit -a -m "local modification" > /dev/null 2>&1
+	git checkout -b deploy-branch > /dev/null 2>&1
+	echo '1' > version.txt
+	git add -A .
+	git commit -m 'branch modification' > /dev/null 2>&1
+	$GIT_FTP pull > /dev/null 2>&1
+	rtrn=$?
+	assertEquals 0 $rtrn
+	assertTrue ' external file not downloaded' "[ -r 'external.txt' ]"
+	assertTrue ' version.txt of deploy-branch not found' "[ -r 'version.txt' ]"
+	assertEquals '## deploy-branch' "$(git status -sb)"
+}
+
+test_pull_no_commit() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' > external.txt
+	curl -T external.txt $CURL_URL/ 2> /dev/null
+	rm external.txt
+	echo 'own content' > internal.txt
+	git add . > /dev/null 2>&1
+	git commit -a -m "local modification" > /dev/null 2>&1
+	LOCAL_SHA1=$(git log -n 1 --pretty=format:%H)
+	$GIT_FTP pull --no-commit > /dev/null 2>&1
+	rtrn=$?
+	assertEquals 0 $rtrn
+	assertTrue ' external file not downloaded' "[ -r 'external.txt' ]"
+	assertEquals $LOCAL_SHA1 $(git log -n 1 --pretty=format:%H)
+}
+
+test_pull_dry_run() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	echo 'own content' > internal.txt
+	git add . > /dev/null 2>&1
+	git commit -a -m "local modification" > /dev/null 2>&1
+	pull=$($GIT_FTP pull --dry-run 2> /dev/null)
+	assertEquals 0 $?
+	assertTrue ' external file downloaded' "[ ! -e 'external.txt' ]"
+	assertFalse "$pull" "echo \"$pull\" | grep 'Last deployment changed to '"
+	# TODO: idea: really download files and show `git diff` and `git diff --stat`, then reset
+}
+
+test_pull_untracked() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	echo 'own content' > internal.txt
+	echo 'internal.txt' >> .gitignore
+	git add . > /dev/null 2>&1
+	git commit -a -m "ignore some file" > /dev/null 2>&1
+	pull=$($GIT_FTP pull 2> /dev/null)
+	assertEquals 0 $?
+	assertTrue 'internal.txt is missing' "[ -f internal.txt ]"
+	assertEquals '' "$(git log -- 'internal.txt')"
+}
+
+test_pull_stash() {
+	skip_without lftp
+	cd $GIT_PROJECT_PATH
+	$GIT_FTP init > /dev/null
+	echo 'foreign content' | curl -T - $CURL_URL/external.txt 2> /dev/null
+	echo 'own content' > internal.txt
+	git stash -u -q
+	pull=$($GIT_FTP pull 2> /dev/null)
+	assertEquals 0 $?
+	stash_count="$(git stash list | wc -l)"
+	stash_count=$((stash_count+0)) # trims whitespaces produced by wc on OSX
+	assertEquals 1 "$stash_count"
+	assertFalse 'internal.txt appeared' "[ -f internal.txt ]"
+	assertEquals '' "$(git log -- 'internal.txt')"
+}
+
 test_submodule() {
 	submodule='sub'
 	file='file.txt'
@@ -678,6 +876,26 @@ assertContains() {
 	assertTrue "Could not find expression: $1\nTested: $2" "echo \"$2\" | grep '$1'"
 }
 
+skip_without() {
+	command -v $1 > /dev/null || startSkipping
+}
+
+# Git for Windows running with Wine doesn't handle unicode well.
+# Unicode filenames don't work on Windows at the moment.
+# - `sort -u` thinks that 'a' equals 'ä' and therefore omits 'ä'.
+# - `curl` fails to open files with unicode in their name bagder/curl#345
+supports_unicode() {
+	count="$(printf 'a\nä\n' | sort -u | wc -l)"
+	count=$((count+0)) # trims whitespaces produced by wc on OSX
+	test "$count" = "2"
+}
+
+abs_dirname() {
+	# Some systems don't know the `realpath` command and `readlink -f` doesn't work on OSX ...
+	# use workaround from http://stackoverflow.com/a/3572105
+	(cd "$(dirname "$1")" && echo "$PWD")
+}
+
 # load and run shUnit2
-TESTDIR=$(dirname $0)
+TESTDIR="$(abs_dirname "$0")"
 . $TESTDIR/shunit2-2.1.6/src/shunit2
